'''
/** Original work Copyright 2024 damiano IU3QGD
 * https://to-be-defined
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.

 * Addition to the licence: NO company with more than 50employees can use this SW and you are NOT entitled to store this software 
 * on ANY storage that in some way attempts to monetize it (eg: github)
 * corporations and banks are too rich and too powerful, they MUST be split and ALL the money given back to the state

To add tk support in linux

apt-get install python3-tk


Hold classes and functions used to manage the gui

https://www.tcl.tk/

https://wiki.tcl-lang.org/page/custom+cursors


'''
# --------------------------------------------------------------------
# This would be a private class, but python is lacking object isolation
# Note that it is for all purposes a Tk, meaning that this is slightly redundant but may be useful in the future
# @param app_title: a string to display on the top window
# @param _icon_name: the name of a icon file to use for systray and/or app icon
# @param on_closing_fun function to call when the user tries to close
# @param enable_systray if true then enable systray 

from __future__ import annotations

import os
import re
from tkinter import ttk, BooleanVar, StringVar, IntVar
import tkinter
from typing import  Mapping, Tuple
import typing

from glob_fun import frequency_1x_to_Mhz, frequency_1x_to_kHz
import qpystat


table_row_selected_background_color='#adff00'

# =========================================================================
# the root should NOT be hidden, since it would be impossible to get it back
class TkinterRoot(tkinter.Tk):

    # --------------------------------------------------------------------------------------------------------
    # the systray is all to be tested.... so, leave it false !    
    def __init__(self, stat : qpystat.Qpystat, app_title, icon_name, on_closing_fun, enable_systray = False ):  
        tkinter.Tk.__init__(self)

        self.title(app_title)
        
        self.stat=stat
        
        self.app_title = app_title
        self._icon_name = icon_name
        self.on_closing_fun = on_closing_fun

        # systray can be enabled only if icon name is given        
        self.enable_systray=icon_name and enable_systray
        
        self.w_resizable=True

        # I need a list of windows to hide/show when requested        
        self.hide_show_list : list[TkToplevel] = []
#        self.hide_show_list.append(self)

        # does not work on linux
        #self.wm_attributes('-transparentcolor', '#010101')

        self.setRootIcon()
        self.enableSystray()
    
    # --------------------------------------------------------------------------------------------------------
    # just to make it clear what the purpose is, terminate the gui loop
    def gui_destroy(self):
        self.destroy()
    

    # ----------------------------------------------------------------------
    # attempt enable systray if it is what is requested
    # if no systray docking is requested add the required on_closing fun
    # the idea is that when the window is closing it is equivalent of writing quit
    # this is VERY important since it allows to save windows position before win destroy
    # If nothing is defined I define a self way to close the gui, so, it is in the right thread
    def enableSystray(self):
        if self.on_closing_fun:
            self.protocol("WM_DELETE_WINDOW", self.on_closing_fun)
        else:
            self.protocol("WM_DELETE_WINDOW", self.gui_destroy)
        
    # ---------------------------------------------------------------------
    # When on systray mode and a request to show is posted
    def systray_show_window(self, icon, item):
        
        # the icon must be stop and created again
        icon.stop()
        
        # this has to run on swing thread
        self.after_idle(self.after_idle_deiconify)
    
    # --------------------------------------------------------------------
    # callback on all windows to deiconify
    def after_idle_deiconify(self):
        
        for win in self.hide_show_list: 
            win.deiconify()
    
    # -----------------------------------------------------------------
    # When in systray mode and quit is selected we arrive at this point
    def systray_quit_window(self, icon, item):
        
        # the icon must be stop and created again
        icon.stop()
        
        if self.on_closing_fun:
            self.on_closing_fun()
        else:
            self.destroy()
    
    
    # ----------------------------------------------------------------------
    # This will set a nice icon to the root window and attempt to setup systray if requested
    # note that it is not possible to enable systray if there is no icon
    # To have windows to use this icon for the application you need to add this
    
    # myappid = 'prepperdock.release.v0' # arbitrary string
    # ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

    # apparently, there are quite some compatibility issues
    # the problem with this is that it needs PIL, another library to load...
    # imgicon = PhotoImage(file=self._icon_name)
    # self.call('wm', 'iconphoto', self, imgicon)  
    
    # somewhere in main
    def setRootIcon(self):
        
        if not self._icon_name:
            return 
        
        try:
            if os.name == 'nt':
                icon_fname = self._icon_name+'.ico'
                self.wm_iconbitmap(bitmap=icon_fname, default=icon_fname)
            else:
                icon_fname = '@'+self._icon_name+'.xbm'
                self.wm_iconbitmap(bitmap=icon_fname)
                
        except Exception as _exc:
            print("setRootIcon: cannot find ",icon_fname)
    
    # ----------------------------------------------------------------------
    # apparently, there is no way to know if a window is resizable or not 
    def setResizable(self, i_resizable : bool ):
        self.w_resizable=bool(i_resizable)
        # PYTHON how to call superclass methods
        super().resizable(self.w_resizable,self.w_resizable)

    # ----------------------------------------------------------------------
    def isResizable(self) -> bool:
        return self.w_resizable
        
    # ---------------------------------------------------------------------
    # Mostly here to remember how to run a function at later time in swing thread
    # NOTE that you name the function to call and then pass the arguments
    # eg: stat.runGuiIdle(stat.extension.setExtensionStatus,extnum,status)    
    def runIdle(self, func, *args ):
        self.after_idle(func, *args)
        
    # ---------------------------------------------------------------------
    # this is instead when you want to apply a specific delay    
    def runLater(self, ms, func, *args ):
        self.after(ms, func, *args)

    # ---------------------------------------------------------
    # attempt to set the geometry of the window, if the new geometry is bigger then current
    # the parameter is a dictionary where the geometry key should be
    # an example of what comes in is  "geometry": "1089x402+-1206+380" 
    # Also: The current geometry is available ONLY after GUI is UP, so, you have to adjust with a timer, after application is up and running
    # Also: Since this is the Root window it must NOT be possible to hide it on creation, you will not be able to get it back 
    def setGeometry(self, propd ):
        try:
            wg = WindowsGeometry(self.winfo_screenwidth(),self.winfo_screenheight(),self.geometry(),propd['window_geometry'])

            self.geometry(wg.risul_geometry)
                
        except Exception as exc:
            print("TkinterRoot.setGeometry: ",exc)
            
    # ----------------------------------------------------
    # return the equivalent geometry info as a dictionary
    # If given the info will be appended to the given dictionary
    def getGeometry(self) -> typing.Dict[str,str]:

        geodic=dict()
        
        geodic['window_geometry'] = self.geometry()
        geodic['window_state'] = self.state()
        
        return geodic
            

# =========================================================================
# the first two are the width and height
# an example of what comes in is  "geometry": "1089x402+-1206+380" 
# In Tk, a window's position and size on the screen are known as its geometry. A full geometry specification looks like this: widthxheight±x±y.
# Width and height (usually in pixels) are pretty self-explanatory. T
# he x (horizontal position) is specified with a leading plus or minus, so +25 means the left edge of the window should be 25 pixels from the left edge of the screen, 
# while -50 means the right edge of the window should be 50 pixels from the right edge of the screen. 
# Similarly, a y (vertical) position of +10 means the top edge of the window should be ten pixels below the top of the screen, 
# while -100 means the bottom edge of the window should be 100 pixels above the bottom of the screen.

# it is assumed that the desired position is the minimum that can be acceppted

class WindowsGeometry():

    # -----------------------------------------------------------
    def __init__(self, scr_width : int, scr_height : int, cur_geometry : str, want_geometry : str ):

        # assume that the result is the current geometry
        self.risul_geometry = cur_geometry

        # there MAY be a chance that current width and height is stupid, like 1x1+0+0
        # better take care of this and adjust the size, the position can remain 0+0, not an issue
        cur_width,cur_height,_cur_posx,_cur_posy=self.splitGeometry(cur_geometry)
        
        if cur_width < 100:
            cur_width=100
            
        if cur_height < 100:
            cur_height=100
        
        wa_width,wa_height,wa_posx,wa_posy=self.splitGeometry(want_geometry)
        
        if wa_width > cur_width:
            cur_width = wa_width
            
        if wa_height > cur_height:
            cur_height = wa_height
            
        #if cur_posx==0:
        # it actually set it correctly even on withdraw windows but the position goes to 0.0 on show
            
        scr_width = scr_width -20
        scr_height = scr_height -20

        # make sure I am dealing with a reasonably sized monitor, with positive values        
        if scr_width < 100:
            return
            
        # make sure I am dealing with a reasonably sized monitor, with positive values        
        if scr_height < 100:
            return

        # now, generally speaking I should check that the wanted position is kind of reasonably within the size of the screen
        # the logic is that the wanted position should fit within the geometry of the visible space 
        if abs(wa_posx) > scr_width:
            print ('illegal wa_posx '+str(wa_posx)+' scr_width '+str(scr_width))
            return

        if abs(wa_posy) > scr_height:
            print ('illegal wa_posy '+str(wa_posx)+' scr_width '+str(scr_height))
            return
    
        self.risul_geometry="%dx%d+%d+%d" % (cur_width,cur_height,wa_posx,wa_posy)
    
    # -----------------------------------------------------------
    # possibly duplicate, to make sure it is constant behaviour
    def splitGeometry(self, geom : str ):

        # the first blob is the dim_str
        first_array = geom.split('+')
    
        dim_str = first_array[0]
        g_posx = int(first_array[1])
        g_posy= int(first_array[2])
    
        dim_array = dim_str.split('x')
        
        width = int(dim_array[0])
        height = int(dim_array[1])
    
        return width,height,g_posx,g_posy

        
# ==================================================================================
# there should be an interface to declare if a class support GUI hiding and Show
# Window is a Toplevel window, since the GuiRoot should not be possible to be hidden
class GUI_hide_show_window():
    
    # -------------------------------------------------------------------
    # this normally maps to self.withdraw() of a toplevel    
    def GUI_hide_window(self):
        pass
    
    # -------------------------------------------------------------------
    # this normally maps to self.deiconify() of a toplevel    
    def GUI_show_window(self):
        pass

 
# ==================================================================================
# the point is that we always need a title and no window close, otherwise the object disappear
# So, a close will result into a window withdraw that can be reversed later, if needed
class TkToplevel(tkinter.Toplevel):

    # -----------------------------------------------------------
    def __init__(self, title ):
        tkinter.Toplevel.__init__(self)
        
        # we need some sort of title
        self.title(title)

        # this is the way to disable the close button
        self.protocol("WM_DELETE_WINDOW", self.wm_delete_window_fun)

    # -----------------------------------------------------------
    # Called when someone click on the "close window"
    # You can get the object back to display using deiconify()
    def wm_delete_window_fun(self):
        self.withdraw()    

    
    # ---------------------------------------------------------
    # do NOT use the state command, it messes up the position !
    def _setWindowState(self, want_st : str ):
        if want_st == 'normal':
            self.deiconify()
        else:
            self.withdraw()
        
    
    
    # ---------------------------------------------------------
    # attempt to set the geometry of the window
    # the parameter is a dictionary where the geometry key should be
    # an example of what comes in is  "geometry": "1089x402+-1206+380" 
    # Also, note that the current geometry is available ONLY after gui is UP, so, you have to 
    # adjust with a timer, after application is up and running 
    # do NOT use the state command to set window state, it just MESS UP the position !
    # this HAS to be protected in try catch sinceotherwise it will stop the parsing of more settings
    def setGeometry(self, propd ):
        try:
            wg = WindowsGeometry(self.winfo_screenwidth(),self.winfo_screenheight(),self.geometry(),propd['window_geometry'])
    
            self.geometry(wg.risul_geometry)
                
            self._setWindowState(propd['window_state'])
                
        except Exception as exc:
            print("tktoplevel.setGeometry: ",exc)
            
    # ----------------------------------------------------
    # return the equivalent geometry info as a dictionary
    def getGeometry(self) -> typing.Dict[str,str]:

        geodic=dict()
                    
        geodic['window_geometry'] = self.geometry()
        geodic['window_state'] = self.state()

        return geodic
            


# -------------------------------------------------------
# Quite often there is the need to have a confirmation popup
# It just shows the text and an OK button that will close the window        
# use
# import tkinter.messagebox as mbox 
# mbox.showinfo("Attenzione", "La applicazione si chiude dalla schermata principale")
#
"""
class TkPopupMessage(tk.Toplevel):
    def __init__(self, parent, text ):
        tk.Toplevel.__init__(self,parent)
        
        # we need some sort of title
        self.title("Messaggio")
        
        workPanel = ttk.Frame(self)
        
        label=tk.Label(workPanel, text=text, font=('Mistral 18 bold'))
        label.grid(row=0, column=0,padx=(20,20),pady=(30,30))

        okb=ttk.Button(workPanel,text="OK",command=self.ok_click)
        okb.grid(row=1, column=0,padx=(20,20),pady=(0,30))

        workPanel.pack()
        
    def ok_click(self):
        self.destroy()    
        
"""        
        


# --------------------------------------------------------------------
# This class should behave similarly to a Swing BorderLayout
# the way it works is to create one of this passing the proper parent
# then declare how it will be "placed" in the parent with something like this
# panel.pack(fill=tk.BOTH,expand=True)
# NOTE that if you do not fill and expand then it will not resize when needed...
# then you can "add" the pieces you wish, using this logic
# ttk.Label(panel.pnorth, text="NORTH",borderwidth=1,relief="ridge").pack()
# ttk.Label(panel.pleft, text="LEFT",borderwidth=1,relief="ridge").pack()
# ttk.Label(panel.pcenter, text="CENTER",borderwidth=1,relief="ridge").pack()
# ttk.Label(panel.pright, text="RIGHT",borderwidth=1,relief="ridge").pack()
# ttk.Label(panel.psouth, text="SOUTH",borderwidth=1,relief="ridge").pack()
# Clearly, you have to tell to subparts HOW to pack, you may not wish to expand...
        
class BorderLayout(ttk.Frame):
    def __init__(self, container ):
        ttk.Frame.__init__(self, container)
        
        self.panel_list = []
        
        self.pnorth = ttk.Frame(self)
        self.pnorth.pack(fill='x',expand=False)
        self.panel_list.append(self.pnorth)

        # this should be a private variable, python is child crap
        self._fmiddle = ttk.Frame(self)
        self._fmiddle.pack(fill='both',expand=True)

        self.pleft = ttk.Frame(self._fmiddle)
        self.pleft.pack(fill='y',expand=False,side='right')
        self.panel_list.append(self.pleft)
        
        self.pcenter = ttk.Frame(self._fmiddle)
        self.pcenter.pack(fill='both',expand=True,side='right')
        self.panel_list.append(self.pcenter)
        
        self.pright = ttk.Frame(self._fmiddle)
        self.pright.pack(fill='y',expand=False,side='right')
        self.panel_list.append(self.pright)
        
        self.psouth = ttk.Frame(self)
        self.psouth.pack(fill='x',expand=False)
        self.panel_list.append(self.psouth)
        
    # --------------------------------------------------
    # Use as debugging method, show the border of the inner frames, so you can 
    # guess how the layout is happening    
    def showBorder(self):
        for panel in self.panel_list:
            panel['borderwidth'] = 1
            panel['relief'] = 'ridge'
        
                    
                    
                    
# -----------------------------------------------------------
# Wrap the image handling all in this class
# NOTE: this is a tkinter.Button since I need to easily hide border                    
class ImageButton(tkinter.Button):
    def __init__(self, container, an_image : tkinter.PhotoImage, **params):

        #cursor_fname = '@'+str(getResourcesFname('hand-cursor.xbm'))

        # AHHHHH dammed snakes, it MUST be an instance variable since otherwise it get out of scope and it disappear!!!
        # subsample returns one pixel every the number given of pixels, so, 2 actually half the image
        self.btn_ico = an_image
        tkinter.Button.__init__(self, container, image=self.btn_ico, bd=1, relief='flat', cursor='hand2', **params )    
    
    # ------------------------------------------------------------
    # MUST be called from swing and with a good image
    def setImage(self, an_image : tkinter.PhotoImage ):
        # AH, DAMMED, I did forget this !
        self.btn_ico = an_image
        # because, otherwise, it gets garbage collected !
        self.config(image=self.btn_ico)
        



# =================================================================================
# Need to wrap the checkbox into something usable...
# it has to have a name and a simple boolean value..    
# NOTE that the layout is done at caller level, NOT here
class SolarCheckbox(ttk.Checkbutton):    
    
    # -------------------------------------------------------------------------------
    # @parent the parent panel where to bind this checkbox
    # @param cname the name to be used to identify the checkbox, can be None if not used
    # @param text the text to display to the user
    def __init__(self, parent, text, cname=None ):

        # I need the text later
        self.text=text

        # need this to then match it, it can be an id, may be needed
        self.name=cname
        
        # this value can be obtained outside the swing thread
        self.cvalBoolean = False
        
        # this is needed since this is the logic used to propagate state in tkinter...
        # NOTE that I do NOT use the name since it may conflict if it is the same for different check
        self.cval = BooleanVar()
        
        # and this defines that a callback will be called when the value changes
        self.cval.trace_add("write", self._cboxCallback)
        
        # create the checkbox and bind it to the above cval
        ttk.Checkbutton.__init__(self, parent, text=text, onvalue=True, offvalue=False, variable=self.cval)     
        
    def _cboxCallback(self, *args):
        self.cvalBoolean = self.cval.get()
            
    def isChecked(self) -> bool:
        return self.cvalBoolean    
        
    # --------------------------------------------------------------------
    # if you pass an empty or illegal value nothing will change
    def setChecked(self, abol ):
        try:
            self.cval.set(abol)
        except:
            pass

# ==================================================================================

class TV_Entry(ttk.Entry):
    # -------------------------------------------------------------------------------
    # @parent the parent panel where to bind this fentry
    # @param the text variable to bind to
    def __init__(self, parent, textvar, **kwargs ):

        ttk.Entry.__init__(self, parent, textvariable=textvar, **kwargs)

        afont=ttk.Style().lookup("TEntry", "font")

        if afont:
            self['font'] = afont
        

class TV_Entry_UM(ttk.Entry):
    # -------------------------------------------------------------------------------
    # @parent the parent panel where to bind this fentry
    # @param the text variable to bind to
    # @param the text to put as unit measure
    def __init__(self, parent, textvar, um_text : str, **kwargs ):

        ttk.Entry.__init__(self, parent, textvariable=textvar, **kwargs)

        afont=ttk.Style().lookup("TEntry", "font")

        if afont:
            self['font'] = afont



# ==========================================================================================
# vedere https://tkdocs.com/tutorial/widgets.html#entry
# https://www.tcl.tk/man/tcl8.5/TkCmd/ttk_entry.htm
# %V The validation condition that triggered the callback (key, focusin, focusout, or forced).
# %P In prevalidation, the new value of the entry if the edit is accepted. In revalidation, the current value of the entry.
# %d Type of action: 1 for insert prevalidation, 0 for delete prevalidation, or -1 for revalidation.
# I want to have only the int interface and hide all the crap, so, when the content changes (focus out)
# the current value will be parsed and converted to proper int value
class TV_EntryFrequency_MHz_1x(ttk.Frame):
    # -------------------------------------------------------------------------------
    # @parent the parent panel where to bind this fentry
    # @param the text variable to bind to
    def __init__(self, parent_panel : ttk.Frame, a_intvar : tkinter.IntVar, **kwargs ):
        ttk.Frame.__init__(self, parent_panel)
        
        self._entry_intvar=a_intvar

        # needed to push from intvar set to the entry
        self._entry_intvar.trace_add('write', self._onIntvarChanged)

        # this is the actual edited value, a string, initial value is from the provided intvar
        self._entry_editing = tkinter.StringVar(self,value=frequency_1x_to_Mhz(self._entry_intvar.get()))
        
        # this is needed to arrive to the validator function
        self.v_command = (parent_panel.register(self._onValidate), '%P', '%V')
        
        # ok let me create the entry
        self._entry_Mhz = ttk.Entry(self, width=11, textvariable=self._entry_editing, validate='all', validatecommand=self.v_command)
        self._entry_Mhz.pack(side='left')

        ttk.Label(self,text='MHz',style='UMisura.TLabel').pack(side='left')

        afont=ttk.Style().lookup("TEntry", "font")

        if afont:
            self._entry_Mhz['font'] = afont


    # ---------------------------------------------------------------
    # this is for when the interface var changes, easy
    def _onIntvarChanged(self, *args):
        self._entry_editing.set(frequency_1x_to_Mhz(self._entry_intvar.get()))
        
    # --------------------------------------------------------------
    # should split the value and reassemble it
    def _onFocusOut(self, v_val : str ) ->bool:
        v_split=v_val.split('.')

        if v_split[0]:            
            mHz = int(v_split[0])
        else:
            mHz = 446

        f_decs = 0
        
        if len(v_split) > 1:
            f_decs=self._mHz_dec_to_F_1x(v_split[1])
            
        total = mHz*100000+f_decs
        
        self._entry_intvar.set(total)
            
        return True

    # ------------------------------------------------------------------
    # the incoming string is the decimal part of a Mhz frequency, missing one "decimal" (five digits)
    # so, max val is 99999 and min val is 00000 and this is the range that must be returned
    # the input CAN be shorter than five digits and it is assumed padded right with 0 to five digits
    # if there are zeros on the left, the result .... is fine, .... no ?
    # So, the difficult part is to "pad" to arrive to five digits
    def _mHz_dec_to_F_1x(self, v_in : str ) -> int:
        
        if not v_in:
            return 0
        
        s_len=len(v_in)
        o_val = int(v_in)
        
        while s_len < 5:
            o_val = o_val * 10
            s_len += 1
        
        return o_val

    # ------------------------------------------------------------------
    # https://docs.python.org/3/library/re.html
    # I want some digits, one dot and some digits        
    # the {0,5} means repeat the previous "rule" zero to five times
    # the ^ means from the beginning of line
    # the $ means to the end of line
    def _onKeyTyped(self, v_val : str ) -> bool:
        
        # copy paste need to clear the value
        if not v_val:
            return True

        # handling is different if there is a dot or not
        v_split=v_val.split('.')

        if len(v_split) == 1:
            # if NO dot is present, accept a number of three digit max
            return bool(re.search("^[0-9]{1,3}$", v_val))
        elif len(v_split) == 2:
            # if ONE dot is present, accept a number of three digit , a dot, and a number up to five digits
            return bool(re.search("^[0-9]{0,3}\\.[0-9]{0,5}$", v_val))

        #v_risul = mHz * 100000 + self._mHz_dec_to_F_1x(vHz)
        #print(' v_risul ',v_risul)

        # nothing else is allowed
        return False 
    
    # ---------------------------------------------------------------
    # this gets called every time enter, type, exit in input box
    # n+val = the new value of the entry if the edit is accepted. In revalidation, the current value of the entry.
    # operation = (key, focusin, focusout, or forced)
    # on focus out I should update the shared integer value
    def _onValidate(self, n_val, operation) -> bool:
        if operation == 'focusout':
            return self._onFocusOut(n_val)
        elif operation== 'key':
            return self._onKeyTyped(n_val)
        #else:
        #    print("_onValidate ",n_val,' op ',operation)
        
        return True


# ===============================================================================
# A frame that has PACK layout as defauult layout, with specific methods   
# this pack is always TOP, meaning that panels are stacked on top of each other  
class PackPanelTop(ttk.Frame):
    # ---------------------------------------------------------------------------
    def __init__(self, parent, **params):
        ttk.Frame.__init__(self, parent, **params)

        self.p_pady = 2

    # ---------------------------------------------------------------------------
    # I probably need to mess it up in a different way to support passing override values
    # generally speaking, I wish for it to fill the whole x
    def addItem(self, jcomponent, **pad_params):

        if 'fill' in pad_params:
            jcomponent.pack(side='top', pady=self.p_pady,  **pad_params)
        else:
            jcomponent.pack(side='top' ,fill='x' ,pady=self.p_pady,  **pad_params)
            
        
# ===============================================================================
# A frame that has PACK layout as defauult layout, with specific methods     
# this always pack left
class PackPanelLeft(ttk.Frame):
    
    # ---------------------------------------------------------------------------
    def __init__(self, parent, **params):
        ttk.Frame.__init__(self, parent, **params)

        self.p_padx=2

    # ---------------------------------------------------------------------------
    # I probably need to mess it up in a different way to support passing override values
    def addItem(self, jcomponent, **pad_params):

        jcomponent.pack(side='left', padx=self.p_padx, **pad_params)


        
# ===============================================================================
# A frame that has grid layout as defauult layout, with specific methods     
# since this crap language cannot handle post increment
# idiots, reinventing the wheel as an hexagon   
class GridPanel(ttk.Frame):
    
    # ---------------------------------------------------------------------------
    def __init__(self, parent, **params):
        ttk.Frame.__init__(self, parent, **params)
        
        self._current_row=0;
        self._current_column=0;
        self._current_sticky='nsew'
        self.g_padx = 2
        self.g_pady = 1
        
    # ---------------------------------------------------------------------------
    # Call this when you wish to go to the next row
    def nextRow(self):
        self._current_row += 1
        self._current_column = 0
        
    # ---------------------------------------------------------------------------
    # If you ever need to go to the next column, without adding a component
    def nextColumn(self, how_many=1):
        self._current_column += how_many
                        
    # ---------------------------------------------------------------------------
    # this has the logic to allow adding / overriding the default values !
    # if you need an item to span multiple columns, use columnspan=2
    # you need to align right, use sticky='e'
    def addItem(self, jcomponent, **gridparams):

        if not 'sticky' in gridparams:
            gridparams['sticky']=self._current_sticky
            
        jcomponent.grid(row=self._current_row, column=self._current_column, **gridparams)
            
        if self.g_padx:
            jcomponent.grid(padx=self.g_padx)
            
        if self.g_pady:
            jcomponent.grid(pady=self.g_pady)

        self._current_column += 1
        
        
        
# ========================================================================
# This is logically the same as a Java interface, minus some type enforcement 
# Classes should implement this so they can be handled as a eeprom event listener

class EE_events_listener ():

    # ----------------------------------------------------------------
    def __init__(self ):
        pass
    
    # -------------------------------------------------------------------
    # This is called by all registered listeners for EEPROM read
    # The subclass must pick up the data it wishes and updated the gui
    # So, the method picks the block it wants and uses it !
    def updateGuiFromEeprom(self):
        pass
        
    # -------------------------------------------------------------------
    # MUST be overriden by children
    # Update the block to be sent to radio getting data from the GUI
    # It is duty of the method to mark the EEPROM block as dirty, if updated    
    # the method does not return anything since the result is stored in the block
    def update_EEblock_from_GUI(self):
        pass
        
        
# ====================================================================
# Mostly to wrap some idiotic behaviour        
class SolarCombo(ttk.Combobox):
    
    # ----------------------------------------------------------------
    def __init__(self, parent_frame : ttk.Frame, **varargs ):

        self._string_var = StringVar(parent_frame)
        
        ttk.Combobox.__init__(self, parent_frame, textvariable=self._string_var, **varargs)
        
    # ----------------------------------------------------------------
    # CAREFUL: the callback signature MUST be like the following one !
    # def _on_serial_combo_change(self, _v1, _v2, _v3):
    def setWriteCallback(self, a_fun ):
        self._string_var.trace('w', a_fun)
        
        
    def getStringVarValue(self):
        return self._string_var.get()
        
    # ----------------------------------------------------------------
    # this is not the best solution, but it is something
    def getSelectedIndex(self):
        c_index = self.current()
        
        if c_index < 0:
            c_index = 0
            
        return c_index
        
# ======================================================================================
# I am mostly interested on the current index or setting the current index
# I probably need to mess with it     
# Good candidate to be a generic component   
class IndexedCombo(ttk.Combobox):
    
    # ---------------------------------------------------------------------------------
    def __init__(self, parent_panel : ttk.Frame, link_index : IntVar,  **varargs ):
        
        if not 'width' in varargs:
            varargs['width']=5
        
        ttk.Combobox.__init__(self, parent_panel, **varargs)
        
        self._link_idex = link_index
        
        self.bind("<<ComboboxSelected>>", self._onComboBoxSelected)
        
        self._link_idex.trace_add('write', self._onLinkIndexChange)
        
    # ---------------------------------------------------------------------------------
    def _onLinkIndexChange(self, *args):
        try:
            self.current(self._link_idex.get())
        except:
            pass
        
    # ---------------------------------------------------------------------------------
    def _onComboBoxSelected(self, event ):
        self._link_idex.set(self.current())
        
        
            

        
# --------------------------------------------------------------------------------------------------
# The idea here is to have a log window that has a scrollable part, up to a certain number of rows
# stat is the general holder of application vars, container is where to put this log
# NOTE: you MUST apply the proper layout of THIS class on the caller side
# Since python is shy on making things private, I just make this class a Frame
# Not only, but since you have to layout on parent you eventually end up on forwarding quite a few useless methods
# so, keep this to extend ttk.Frame, this is the only reasonable thing to do with python
# Note that since the printing is done in the swing thread, there is no need to hold a lock
class LogPanel(ttk.Frame):
    
    # ------------------------------------------------------------------------
    # You can decide the max number of lines to show
    # to pad contente you can use padding="3p.3p" or you can set padx=3, pady=3
    def __init__(self, stat : qpystat.Qpystat, parent_panel, name : str, maxlines=5000):
        ttk.Frame.__init__(self, master=parent_panel, padding=3)     

        self._stat=stat
        self._maxlines=maxlines

        self._checkPaused : SolarCheckbox;
        
        # Create label inside the frame
        self._newNorthPanel(self, name).pack(fill='x',pady=2)

        # create a text inside the frame
        self._newTextPanel(self).pack(fill='both',expand=True) 
        
    # -----------------------------------------------------------
    # Mostly a reminder that this method is available everywhere there is a 'frame'   
    def runOnGuiIdle(self, func, *args ):
        self.after_idle(func, *args)
        
    # ------------------------------------------------------------
    # create a new north panel
    # The layout of this panel is done at PARENT level
    def _newNorthPanel(self,parent, name):
        apanel = PackPanelLeft(parent)

        # Create label inside the apanel
        # font=("Courier", 14),
        apanel.addItem(ttk.Label(apanel, style='Log.TLabel', text=name))

        # add a checkbox to pause the writing
        self._checkPaused=SolarCheckbox(apanel,'Pause')
        apanel.addItem(self._checkPaused)
    
        apanel.addItem(ttk.Button(apanel,text="Clear" , command=self._clickClear ))
    
        return apanel

    # --------------------------------------------------------------
    # When button is click, clear, it is thread safe since is in gui thread
    def _clickClear(self):
        self.text.delete('1.0', 'end')
        
    # ------------------------------------------------------------
    # this creates a new panel to hold both the text and scrollbar    
    def _newTextPanel(self,parent_panel):
        # The layout of this apanel is done at caller level
        apanel = ttk.Frame(parent_panel)
        
        # Create text widget and specify size.
        self.text = tkinter.Text(apanel, height = 10, width = 80)
        self.text.pack(fill='both',expand=True,side='left')
        
        scroll_bar = ttk.Scrollbar(apanel,orient='vertical',command=self.text.yview)
        scroll_bar.pack(fill='y', expand=False, side='left')
        
        self.text['yscrollcommand'] = scroll_bar.set
        
        self.setWrap(False)
                
        return apanel
        
        
    # --------------------------------------------------------------    
    def getTextString(self) -> str :
        astring = self.text.get("1.0",'end-1c')    
        return astring
        
    # -------------------------------------------------------------
    # true, enables wrap, false nowrap
    def setWrap (self, enabled):
        if enabled:
            self.text.config(wrap='word')    
        else:
            self.text.config(wrap='none')
        
    # --------------------------------------------------------------------
    # Use this one to actually print something
    # @param addnl if true a newline will be added after the given text
    def _println_gui (self, msg : str, addnl : bool ):

        if self._checkPaused.isChecked():
            return

        if not self.text.winfo_exists():
            # this may happen if in closing mode and messages want to be logged
            # if the widget is destroyed then print into the standard out
            # NOTE this will NOT happen if you call the window.destroy() of the main GUI from a gui idle callback
            print('logui wnot: ',msg)
            return 
        
        endpos=self.text.index('end')
        split=endpos.split('.')
        lines = int(split[0])
        
        if lines > self._maxlines :
            delindex=lines - self._maxlines 
            self.text.delete('1.0', "%d.end" % delindex)
            
        self.text.insert(tkinter.END, msg)
        
        if addnl:
            self.text.insert(tkinter.END, "\n")
        
        self.text.see(tkinter.END)

    # -------------------------------------------------------------------------
    # Always run the print on GUI idle, this solves a lot of issues on threading
    def println(self,msg):
        if len(msg) > 0:
            self.runOnGuiIdle(self._println_gui,msg,True)

    # -------------------------------------------------------------------------
    # this does NOT add a newline at the end
    def write(self,msg):
        if len(msg) > 0:
            self.runOnGuiIdle(self._println_gui,msg, False)

            
# ========================================================================
# https://gist.github.com/novel-yet-trivial/3eddfce704db3082e38c84664fc1fdf8/stargazers    
class VerticalScrolledFrame:
    """
    A vertically scrolled Frame that can be treated like any other Frame
    ie it needs a master and layout and it can be a master.
    :width:, :height:, :bg: are passed to the underlying Canvas
    :bg: and all other keyword arguments are passed to the inner Frame
    note that a widget layed out in this frame will have a self.master 3 layers deep,
    (outer Frame, Canvas, inner Frame) so 
    if you subclass this there is no built in way for the children to access it.
    You need to provide the controller separately.
    """
    def __init__(self, master, **kwargs):
        width = kwargs.pop('width', None)
        height = kwargs.pop('height', None)
        bg = kwargs.pop('bg', kwargs.pop('background', None))
        self.outer = ttk.Frame(master, **kwargs)

        self.vsb = tkinter.Scrollbar(self.outer, orient=tkinter.VERTICAL)
        self.vsb.pack(fill=tkinter.Y, side=tkinter.RIGHT)
        self.canvas = tkinter.Canvas(self.outer, highlightthickness=0, width=width, height=height, bg=bg)
        self.canvas.pack(side=tkinter.LEFT, fill=tkinter.BOTH, expand=True)
        self.canvas['yscrollcommand'] = self.vsb.set
        # mouse scroll does not seem to work with just "bind"; You have
        # to use "bind_all". Therefore to use multiple windows you have
        # to bind_all in the current widget
        self.canvas.bind("<Enter>", self._bind_mouse)
        self.canvas.bind("<Leave>", self._unbind_mouse)
        self.vsb['command'] = self.canvas.yview

        self.inner = ttk.Frame(self.canvas)
        # pack the inner Frame into the Canvas with the topleft corner 4 pixels offset
        self.canvas.create_window(4, 4, window=self.inner, anchor='nw')
        self.inner.bind("<Configure>", self._on_frame_configure)

        self.outer_attr = set(dir(tkinter.Widget))

    def __getattr__(self, item):
        if item in self.outer_attr:
            # geometry attributes etc (eg pack, destroy, tkraise) are passed on to self.outer
            return getattr(self.outer, item)
        else:
            # all other attributes (_w, children, etc) are passed to self.inner
            return getattr(self.inner, item)

    def _on_frame_configure(self, event=None):
        _x1, _y1, x2, y2 = self.canvas.bbox("all")
        height = self.canvas.winfo_height()
        self.canvas.config(scrollregion = (0,0, x2, max(y2, height)))

    def _bind_mouse(self, event=None):
        self.canvas.bind_all("<4>", self._on_mousewheel)
        self.canvas.bind_all("<5>", self._on_mousewheel)
        self.canvas.bind_all("<MouseWheel>", self._on_mousewheel)

    def _unbind_mouse(self, event=None):
        self.canvas.unbind_all("<4>")
        self.canvas.unbind_all("<5>")
        self.canvas.unbind_all("<MouseWheel>")

    def _on_mousewheel(self, event):
        """Linux uses event.num; Windows / Mac uses event.delta"""
        if event.num == 4 or event.delta > 0:
            self.canvas.yview_scroll(-1, "units" )
        elif event.num == 5 or event.delta < 0:
            self.canvas.yview_scroll(1, "units" )

    def __str__(self):
        return str(self.outer)
    
    
    
# ==================================================================================
# Column definition is stored into an object since I need to switch the sorting for a given column
# This is used in TreeView to display a table
class SolarTableColdef:

    # ----------------------------------------------------------------
    # Constructor
    def __init__(self, qcol_name : str , col_label : str , isNumeric, width, stretch=False ):
        self.col_name : str = qcol_name
        self.col_label : str =col_label
        self.sort_down=True                # means 4,3,2,1 ...
        self.isNumeric=isNumeric
        self.width=width
        self.stretch=stretch
        
    # --------------------------------------------------
    # this is the equivalent of toString()
    def __str__(self):
        return self.col_name+":"+self.col_label+":"+str(self.sort_down)
    
    
# ==================================================================================
# python table handling has a few quirks that are better wrapped up

class SolarTable():
    
    # ----------------------------------------------------------------------------------
    # You need a parent and a mapping for the columns    
    # Also, since 99% of times, you need a scrollbar, attach it to a panel, as it should be
    # list(qschcols_map.keys()) is needed since just qschcols_map.keys() produce a view of the map and the Treeview does not like it
    # takefocus=0 does not work to prevent the tree to steal focus from the Entry on tree filter
    # https://anzeljg.github.io/rin2/book2/2405/docs/tkinter/ttk-Treeview.html

    def __init__(self, parent_panel : ttk.Frame, col_list : Mapping[str,SolarTableColdef], **varargs ):
        
        self._work_panel = ttk.Frame( parent_panel)

        self._work_panel.grid_rowconfigure(0, weight = 1)
        self._work_panel.grid_columnconfigure(0, weight = 1)
        
        self._col_list : Mapping[str,SolarTableColdef] = col_list
        self._tktable = ttk.Treeview(self._work_panel,columns=list(col_list.keys()), show='headings', **varargs)

        for colu in col_list.values():
            alambda=lambda _col=colu.col_name : self._table_sort_column( _col )
            self._tktable.heading(colu.col_name, text=colu.col_label, anchor='w', command=alambda )
            self._tktable.column(colu.col_name, minwidth=50, width=colu.width, stretch=colu.stretch)

        self._tktable.grid(row=0, column=0, sticky='nsew')
        
        self._scrollbar = ttk.Scrollbar(self._work_panel, orient=tkinter.VERTICAL, command=self._tktable.yview)
        self._tktable.configure(yscrollcommand=self._scrollbar.set)
        
        self._scrollbar.grid(row=0, column=1, sticky='ns')

    # ----------------------------------------------------------------------
    # When you need the inner workings
    def getComponent(self) -> ttk.Frame:
        return self._work_panel

    # ----------------------------------------------------------------------
    # When you need the inner workings
    def getTable(self) -> ttk.Treeview:
        return self._tktable


    # -------------------------------------------------------------------
    # Used to sort a table by a given column
    # it gets all children keys, create a list that gets sorted and then move items around        
    def _table_sort_column(self, tblcol_name ):
        a_coldef : SolarTableColdef = self._col_list[tblcol_name]
        
        # the following returns a tuple of all "keys", eg: ('I001', 'I002', 'I003', 'I004', 'I005'...... 
        # print(self._tktable.get_children())

        # self._tktable.set(key, tblcol_name)
        # returns the value associated with the given col name at the given key

        lista = [(self._tktable.set(key, tblcol_name), key) for key in self._tktable.get_children()]

        # if the column is numeric I need to convert it to int the value     
        # NOTE that I am sorting the VALUE, not the key !   
        if a_coldef.isNumeric:
            lista = [(int(val),key) for val,key in lista]
            
        lista.sort(reverse=a_coldef.sort_down)
    
        # reverse the sorting, for next time
        a_coldef.sort_down = not a_coldef.sort_down 
    
        # rearrange items in sorted positions. enumerate returns the index and the actual value
        for index, (_val, key) in enumerate(lista):
            self._tktable.move(key, '', index)

    # ----------------------------------------------------------------------
    # Clear the whole table    
    def clear(self):
        self._tktable.focus_set()

        children = self._tktable.get_children()

        for child in children:
            self._tktable.delete(child)
        
    # ----------------------------------------------------------------------
    def row_exist(self, row_id ) -> bool:
        return self._tktable.exists(str(row_id))
        
    # ----------------------------------------------------------------------
    def row_detach(self, row_id):
        self._tktable.detach(str(row_id))
                
    # ----------------------------------------------------------------------
    def row_delete(self, row_id):
        self._tktable.delete(str(row_id))

    # ----------------------------------------------------------------------
    # this will reattach/move the given iid at the beginning
    def row_reattach(self, row_id ):
        # params are the iid, the parent and where to move
        # if move is 0 then it is moved to beginning
        self._tktable.move(str(row_id),'',0)

    # ----------------------------------------------------------------------
    # this is appropriate for a table
    # I am not using the "index" since there is some problem when you want to sort ascending/descending
    def row_append(self, row_id, row_values ):
        self._tktable.insert(parent='',index='end',iid=str(row_id),values=row_values)
            
    # ------------------------------------------------------------------------
    # NOTE that the iid is really anything you wish !
    # @return the array of selected iid, can be multiple iid            
    def getSelected_iid(self) -> Tuple[str, ...]:
        return self._tktable.selection()
        
    def set_row(self, row_id, v_values, v_tags ):
        self._tktable.item(str(row_id), values=v_values, tags=v_tags)

    def see_row(self, row_id):
        self._tktable.see(str(row_id))
        
    def focus_row(self, row_id):
        self._tktable.focus(str(row_id))
        
        




















# ==========================================================================================
# For comments, see the MHz version !
class TV_EntryFrequency_kHz_1x(ttk.Frame):
    # -------------------------------------------------------------------------------
    # @parent the parent panel where to bind this fentry
    # @param the text variable to bind to
    def __init__(self, parent_panel : ttk.Frame, a_intvar : tkinter.IntVar, **kwargs ):
        ttk.Frame.__init__(self, parent_panel)
        
        self._entry_intvar=a_intvar

        # needed to push from intvar set to the entry
        self._entry_intvar.trace_add('write', self._onIntvarChanged)

        # this is the actual edited value, a string, initial value from provided var
        self._entry_editing = tkinter.StringVar(self,value=frequency_1x_to_kHz(self._entry_intvar.get()))
        
        # this is needed to arrive to the validator function
        self.v_command = (parent_panel.register(self._onValidate), '%P', '%V')
        
        # ok let me create the entry
        self._entry_kHz = ttk.Entry(self, width=11, textvariable=self._entry_editing, validate='all', validatecommand=self.v_command)
        self._entry_kHz.pack(side='left')

        ttk.Label(self,text='kHz',style='UMisura.TLabel').pack(side='left')

        afont=ttk.Style().lookup("TEntry", "font")

        if afont:
            self._entry_kHz['font'] = afont


    # ---------------------------------------------------------------
    # this is for when the interface var changes, easy
    def _onIntvarChanged(self, *args):
        self._entry_editing.set(frequency_1x_to_kHz(self._entry_intvar.get()))
        
    # --------------------------------------------------------------
    # should split the value and reassemble it
    def _onFocusOut(self, v_val : str ) ->bool:
        v_split=v_val.split('.')

        if v_split[0]:            
            kHz = int(v_split[0])
        else:
            kHz = 0

        f_decs = 0
        
        if len(v_split) > 1:
            f_decs=self._kHz_dec_to_F_1x(v_split[1])
            
        total = kHz*100+f_decs
        
        self._entry_intvar.set(total)
            
        return True

    # ------------------------------------------------------------------
    def _kHz_dec_to_F_1x(self, v_in : str ) -> int:
        
        if not v_in:
            return 0
        
        s_len=len(v_in)
        o_val = int(v_in)
        
        while s_len < 2:
            o_val = o_val * 10
            s_len += 1
        
        return o_val

    # ------------------------------------------------------------------
    def _onKeyTyped(self, v_val : str ) -> bool:
        
        if not v_val:
            return True

        v_split=v_val.split('.')

        if len(v_split) == 1:
            return bool(re.search("^[0-9]{1,4}$", v_val))
        elif len(v_split) == 2:
            # if ONE dot is present, accept a number of three digit , a dot, and a number up to TWO digits
            return bool(re.search("^[0-9]{0,4}\\.[0-9]{0,2}$", v_val))

        # nothing else is allowed
        return False 
    
    # ---------------------------------------------------------------
    def _onValidate(self, n_val, operation) -> bool:
        if operation == 'focusout':
            return self._onFocusOut(n_val)
        elif operation== 'key':
            return self._onKeyTyped(n_val)
        
        return True

# ====================================================================
# mostly to have a decent name 
class GuiTabPanel(ttk.Notebook):
    
    # ----------------------------------------------------------------
    def __init__(self, parent, **varargs):
        ttk.Notebook.__init__(self, parent, **varargs)

    # ----------------------------------------------------------------
    def addItem(self, jcomponent, tab_name : str ):
        super().add(jcomponent, text=tab_name)




    