'''
/** Original work Copyright 2024 damiano IU3QGD
 * https://to-be-defined
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 *     Unless required by applicable law or agreed to in writing, software
 *     distributed under the License is distributed on an "AS IS" BASIS,
 *     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *     See the License for the specific language governing permissions and
 *     limitations under the License.
 
 * Addition to the licence: NO company with more than 50employees can use this SW and you are NOT entitled to store this software 
 * on ANY storage that in some way attempts to monetize it (eg: github)
 * corporations and banks are too rich and too powerful, they MUST be split and ALL the money given back to the state
 
 

Other parts of the system should send messages here to be sent
There has to be a class that manage the queue of incoming commands
this is because the GUI cannot do IO (will be come sluggish) and the serial channel is NON sharable by default

'''


from __future__ import annotations

import pathlib
from queue import Queue
import struct
from threading import Thread
import time 
from tkinter import StringVar
import tkinter
from tkinter.filedialog import askopenfilename, asksaveasfilename
import traceback
from typing import cast
import typing

import serial.tools.list_ports  # type: ignore

from app_config import AppConfig, ConfigSavable
import glob_eeprom
from glob_fun import getApplicationResourcesPath
from glob_gui import TkToplevel, TV_Entry, LogPanel, SolarCombo, \
    GUI_hide_show_window, PackPanelTop, GridPanel, GuiTabPanel
from glob_ipp import IPPU_pollpending_reqid, Qsk5_res_RadioStatus, IppChunk, \
    Qsk5_res_Counters, Qsk5_res_Spectrum, Qsk5_res_Loudest, Qsk5_res_PollDtmf, \
    Qsk5_res_RadioPrintf, Qsk5_res_RadioIdentity, Qsk5_res_RadioConfiguration, \
    Qsk5_res_Generic, Qsk5_res_Eeprom, Qsk5_req_read_eeprom_quansheng
import glob_ipp
from ipp_parser import Ippparser
from qcounters_gui import Qcounters_gui
import qpystat
import tkinter.ttk as ttk
from uvk5_lib import RadioReceiveRes, UVK5_FLASH_BLOCK_LEN, \
    UVK5_FLASH_MAX_SIZE, Libuvk5serial, QuaFlash, Flash_V2, \
    UVk5_BLOADER_BEACONv2_hex, UVk5_BLOADER_BEACONv5_hex, Flash_V5, \
    BloaderBeaconParser


# ====================================================================
class QconnectThread (Thread,ConfigSavable,GUI_hide_show_window):

    # ----------------------------------------------------------------
    def __init__(self, stat : qpystat.Qpystat):
        Thread.__init__(self,name="QConnect THREAD",daemon=True)
        
        self.stat = stat
        
        self._oqueue : Queue = Queue()
        self.uvk5fw = Uvk5Firmware(stat)
        
        self.stat.appconfig.addMyselfToSavables(self)

        self._toplevel = TkToplevel("QConnect Window")
        
        a_panel=PackPanelTop(self._toplevel, padding=2 )     
    
        a_panel.addItem(self._newTopButtonsPanel(a_panel))
        a_panel.addItem(self._newConfigPanel(a_panel))
        a_panel.addItem(self._newStatusPanel(a_panel))
        a_panel.addItem(self._newRssiPanel(a_panel))
        a_panel.addItem(self._newTabbedPanel(a_panel),fill='both',expand=True)
        
        a_panel.pack(fill='both',expand=True)

        self.GUI_hide_window() 
        
        self.cur_radio_fw_version=''
        self.cur_radio_cpu_id_hex=''

        self._fw_chooser_showing=False
        self._fw_have_write_req=False
        self._auto_load_fw=None
        
        # initialized here since it needs the log window
        self.stat.libqsk5 = Libuvk5serial(self._qconn_log)
        
        self._println("QConnect init complete")

    # -------------------------------------------------------------------
    def _newTabbedPanel(self, parent_panel : ttk.Frame ) -> ttk.Notebook:
        a_tabbed = GuiTabPanel(parent_panel)

        # --------------- first component
        self._qconn_log = LogPanel(self.stat,a_tabbed,"Connection Log")
        self._qconn_log.setWrap(True);
        self._qconn_log.pack(fill='both',expand=True)
        
        a_tabbed.addItem(self._qconn_log, "Log")

        # ---------------- second component
        
        self._qcounters = Qcounters_gui(self.stat, a_tabbed)  
        
        a_tabbed.addItem(self._qcounters.getWorkPanel(), "Qradio Counters")
        
        return a_tabbed


    # -------------------------------------------------------------------
    # prints something on this window log
    def _println(self, msg):
        self._qconn_log.println(msg)    
        
    # --------------------------------------------------------------------
    # show a window that has been hidden using the withdraw method
    def GUI_show_window(self):
        self._toplevel.deiconify()    

    # --------------------------------------------------------------------
    # hide completely a window
    def GUI_hide_window(self):
        self._toplevel.withdraw()    

    # -----------------------------------------------------------
    # Mostly a reminder that this method is available everywhere    
    def _runOnGuiIdle(self, func, *args ):
        self._toplevel.after_idle(func, *args)

    # -------------------------------------------------------------------------------
    # make a new frame and put an input line in it
    def _newConfigPanel(self, parent):

        panel=ttk.Frame(parent,padding='3')
        
        # this variable will store whatever the user has typed
        self._serial_name_var = StringVar(panel)

        # I have a combo with the list of supposedly detected ports
        ttk.Label(panel,text="Detected ports: ").pack(side='left')
        self._serial_name_combo = SolarCombo(panel)
        self._serial_name_combo.setWriteCallback(self._on_serial_combo_change)

        self._serial_name_combo.pack(side='left')
        
        ttk.Label(panel,text=" Selected port:").pack(side='left')
        TV_Entry(panel, self._serial_name_var, width=12).pack(side='left')
        
        return panel



    # ----------------------------------------------------------------------------------
    # NOTE that there is style definition in _adjustTtkStyle
    def _newRssiPanel(self, parent):

        rssi_panel = GridPanel( parent, borderwidth=1, relief='solid') 

        rssi_panel.columnconfigure(2, weight = 1)   # since the other is zero, this will get all the extra space

        w_font = ('Consolas',10)
        w_font_bold = ('Consolas',10,'bold')

        #RSSI
        rssi_panel.addItem(ttk.Label(rssi_panel, font=w_font, text="RSSI dBm"))
        self._rssi_vLabel = ttk.Label(rssi_panel, font=w_font_bold)
        rssi_panel.addItem(self._rssi_vLabel,sticky='e')
        self._rssiBar = ttk.Progressbar(rssi_panel, orient='horizontal', mode='determinate', maximum=0x1FF/2)
        rssi_panel.addItem(self._rssiBar)

        rssi_panel.nextRow()

        #NOISE
        rssi_panel.addItem(ttk.Label(rssi_panel, font=w_font, text="Ex-noise"))
        self._noise_Label = ttk.Label(rssi_panel, font=w_font_bold)
        rssi_panel.addItem(self._noise_Label,sticky='e')
        self.noiseBar = ttk.Progressbar(rssi_panel, orient='horizontal', mode='determinate', maximum=0x7F)
        rssi_panel.addItem(self.noiseBar)

        rssi_panel.nextRow()

        #GLITCH
        rssi_panel.addItem(ttk.Label(rssi_panel, font=w_font, text="Glitch indicator   "))
        self._glitch_Label = ttk.Label(rssi_panel, font=w_font_bold)
        rssi_panel.addItem(self._glitch_Label,sticky='e')
        self.glitchBar = ttk.Progressbar(rssi_panel, orient='horizontal', mode='determinate', maximum=0xFF)
        rssi_panel.addItem(self.glitchBar)

        rssi_panel.nextRow()
        
        #BATTERY
        rssi_panel.addItem(ttk.Label(rssi_panel, font=w_font, text="Battery Voltage"))
        self._battV_Label = ttk.Label(rssi_panel, font=w_font_bold)
        rssi_panel.addItem(self._battV_Label,sticky='e')
        self.battBar = ttk.Progressbar(rssi_panel, orient='horizontal', mode='determinate', maximum=1000)
        rssi_panel.addItem(self.battBar)

        rssi_panel.nextRow()

        #CPU Temperature
        rssi_panel.addItem(ttk.Label(rssi_panel, font=w_font, text="CPU Temperature"))
        self._tcpu_Label = ttk.Label(rssi_panel, font=w_font_bold)
        rssi_panel.addItem(self._tcpu_Label,sticky='e')
        self.tcpuBar = ttk.Progressbar(rssi_panel, orient='horizontal', mode='determinate', maximum=4000)
        rssi_panel.addItem(self.tcpuBar)

        return rssi_panel 
    
    # ---------------------------------------------------------------------------------
    # this updates only the chunk with radio status
    def _updateRadioStatus(self, chunk : Qsk5_res_RadioStatus):
        
        rssi = chunk.rx_rssi / 2 - 160
        
        self._rssi_vLabel.config(text="{:7.2f}".format(rssi))
        self._rssiBar['value'] = rssi+160
        
        self._noise_Label.config(text="{:d}".format(chunk.rx_noise))
        self.noiseBar['value'] = chunk.rx_noise
                
        self._glitch_Label.config(text="{:d}".format(chunk.rx_glitch))
        self.glitchBar['value'] = chunk.rx_glitch

        self._battV_Label.config(text="{:.2f}".format(chunk.batteryV))
        self.battBar['value'] = chunk.batteryV_2dec

        self._tcpu_Label.config(text="{:.2f}".format(chunk.tc_cpu))
        self.tcpuBar['value'] = chunk.tc_cpu_2dec

        self.rxFrequency.config(text="RX Frequency HZ  {}".format(chunk.rxfreq))
        self.radioStatus.config(text="Radio Status bits  {:#04x}".format(chunk.radio_status))
    

    # -------------------------------------------------------------------------------
    # make a new frame and put an input line in it
    def _newStatusPanel(self, parent):

        panel=ttk.Frame(parent, padding=3,borderwidth=1, relief='solid')
        
        self.rxFrequency = ttk.Label(panel, justify=tkinter.LEFT,font=('calibri',14))
        self.rxFrequency.pack(side='top',fill='x',expand=True);
        
        self.radioStatus = ttk.Label(panel, justify=tkinter.LEFT,font=('calibri',14))
        self.radioStatus.pack(side='top',fill='x',expand=True);
        
        return panel


    # -------------------------------------------------------------------------------
    # make a new frame and put buttons into it, but stay in this class
    def _newTopButtonsPanel(self, container):

        cframe = ttk.Frame(container, padding=3) # borderwidth=3, relief='ridge')

        ttk.Button(cframe,text="Save Log" , command=self._clicSaveLog ).pack(side='right',padx=3)
        ttk.Button(cframe,text="Write Firmware" , command=self._clicWriteFirmware ).pack(side='right',padx=3)

        return cframe
    
    # ---------------------------------------------------------------------------------------
    # this MUST be run in swing thread AND it should NOT be queued AT ALL if _fw_chooser_showing is true !
    def _clicWriteFirmware(self):

        if self._fw_chooser_showing:
            return
        
        if self._fw_have_write_req:
            return
        
        self._fw_chooser_showing = True
        
        # show the Qconnect window, otherwise the sheeple will get scared
        self.GUI_show_window()
        
        # then show the file chooser
        self._fw_load_filename()

        self._fw_chooser_showing = False
    
    # --------------------------------------------------------------------------------------
    def _fw_choose_filename(self, initdir : pathlib.Path ) -> pathlib.Path:
        filetypes = ( ('binary files', '*.bin'),   ('All files', '*.*')   )

        filename = askopenfilename(parent=self._toplevel, title='Open FIRMWARE file', initialdir=initdir, filetypes=filetypes)

        self._println("_clicWriteFirmware: file "+str(filename))
    
        return pathlib.Path(str(filename))
    
    # --------------------------------------------------------------------------------------
    # The actual writing HAS to be done using the polling thread
    # First I need to pick up the file
    def _fw_load_filename(self):

        initdir : pathlib.Path=getApplicationResourcesPath()
        
        if self._auto_load_fw:
            filename = initdir / 'qsk5_prepper_fw.bin'
        else:
            filename = self._fw_choose_filename(initdir)

        # the user simply pressed cancel
        if not filename:
            return

        if self.uvk5fw.load_from_filename(filename):        
            self._println("Firmware size "+str(self.uvk5fw.getFirmwareSize()))
            # ask to poller to write the FW
            self._fw_have_write_req=True 
        else:
            self._println("_clicWriteFirmware: CANNOT load "+str(filename))
    
    # --------------------------------------------------------------------------------------
    # https://docs.python.org/3/library/dialog.html
    def _clicSaveLog(self):
        try:
            
            files = [('Text Document', '.txt'), ('All Files', '*.*')]
            ffname = asksaveasfilename(parent=self._toplevel, title="Save Log to File", initialfile="prepperdock-qsk5-log.txt", filetypes = files, defaultextension = '.txt' ) 

            if not ffname:
                return 
            
            with open(ffname, 'wb') as fileo:
                astring=self._qconn_log.getTextString()
                fileo.write(astring.encode('UTF-8', 'replace'))
                
                self._println('--')
                self._println("saved file name "+str(ffname))
                
        except Exception as exc :
            self._println("_clicSaveLog "+str(exc))


    # -------------------------------------------------------------------
    # implement the config savable
    def appImportConfig(self, cfg : AppConfig ):
        adict = cfg.getDictValue("qconnect_cfg", {})
        
        try:
            self._println("geometry "+self._toplevel.geometry())

            self._serial_name_var.set(adict['poll_port'])    
            self._toplevel.setGeometry(adict['gui_cfg'])

            self._auto_load_fw=adict['auto_load_fw']    
        except Exception as _exc :
            pass            

    # -------------------------------------------------------------------
    # implement the config savable
    def appSaveConfig(self, cfg : AppConfig ):
        # get the current dictionary bound to qconnect
        adict = cfg.getDictValue("qconnect_cfg", {})

        adict['gui_cfg'] = self._toplevel.getGeometry()
        adict['poll_port'] = self._serial_name_var.get()

        cfg.setDictValue("qconnect_cfg", adict )            
    
    # ------------------------------------------------------
    # use this method to queue a command to be executed
    # it has to be a child of a command
    def queuePut (self,line : glob_ipp.Qsk5_command):
        self._oqueue.put(line)

    def SHIT_python_circular_import(self):
        cmd = Qsk5_req_read_eeprom_quansheng(0,glob_eeprom.GlobEeprom.ee_blocks_max)
        self.queuePut(cmd)



    # ----------------------------------------------------------------------------------
    # called when the user changes the combo
    # it is ALSO called when changing the combo selection programmatically !
    # it should change the text value in the input box AND disconnecte the current port
    def _on_serial_combo_change(self, _v1, _v2, _v3):
        
        sel_port = self._serial_name_combo.getStringVarValue()
        
        self._println("combo selected "+sel_port)
        
        self._serial_name_var.set(sel_port)

        # if there was anything ocnnected, disconnect
        self.stat.libqsk5.disconnect()

    # --------------------------------------------------------------
    # attempt to adjust the combo with thelist of serial ports detected
    # NOTE that an emptylist is a VALID setting, meaning, non ports detected
    def _swing_update_ports_combo(self, ports_list : typing.List[str]):
        
        sorted_list = sorted(ports_list)
        current_list = self._serial_name_combo['values']
        
        if sorted_list == current_list:
            return 
         
        self._serial_name_combo['values'] = sorted(ports_list)

        if not sorted_list:
            return 
        
        u_port_val = self._serial_name_var.get()
        
        # the isea is that if the user has NOT set a value AND there is NOTHING selected in the combo
        # this is because a user combo change WILL result in a change AND disconnect !
        if not u_port_val and self._serial_name_combo.current() < 0:
            self._serial_name_combo.current(0)
            
        #self._println('UFFA '+self._serial_name_combo.getStringVarValue())
            


    # --------------------------------------------------------------
    # of 3 strings: port name, human readable description and a hardware ID.
    # this is called when I wish to detect ports, once the ports are detected they will be push to the combo
    def _serial_ports_detect(self):
        port_list=serial.tools.list_ports.comports()

        ports_list = []

        for row  in port_list:
            a_port : typing.Tuple[str,str,str] = row
            self._println('port '+str(a_port[0])+' '+str(a_port[1]) )
            ports_list.append(a_port[0])

        self._toplevel.after_idle(self._swing_update_ports_combo, ports_list)

            


    
    # ------------------------------------------------------
    # attempt to open the currently selected serial port
    # this is NOT in swing thread, so, no real gui messing here
    # TRY to request a combo reconfigure    
    def _open_serial_port(self) -> bool:

        # request a port detect, in any case
        self._serial_ports_detect()

        # I use a combo with detected ports AND an input box, since, I REALLY wish to give the user an option to type what they wish
        portname=self._serial_name_var.get()
        
        if not portname:
            self._println("Qconnect: port name empty")
            return False
    
        self._println("Qconnect opening "+portname)

        # clear up any possibly previous opened stream
        self.stat.libqsk5.disconnect()

        return self.stat.libqsk5.connect(portname)
    
    # ------------------------------------------------------
    # wrapped open, to avoid lots of garbage on open error
    def _open_serial_port_wrap(self) -> bool:
        try:
            return self._open_serial_port()
        except Exception as _err:
            self._println("serial port open FAIL")
            return False
    
    # ------------------------------------------------------
    # Attempt to open the device and do some polling
    def _open_and_poll(self):
        try:
            
            if not self._open_serial_port_wrap():
                return
            
            self._println("  Serial port is OPEN")

            while self._onePoll():
                time.sleep(0.05)
        
        except Exception as _err:
            self._println(traceback.format_exc())
    

    
    # --------------------------------------------------------------
    # Called by the Thread, do NOT call it, use start()
    # I wish this to stop on first error, so I can investigate
    def run(self):
        self._println ("START "+self.name)

        try_count=0
        
        while True:
            self._println ("----------> "+str(try_count))
            self._open_and_poll()
            self._println ("poll errors: wait 5s")
            time.sleep(5)
            try_count+=1

        self._println ("END "+self.name)


    # --------------------------------------------------------------
    # return a byte array of a string padded to the given len
    def _pad_String(self, astring : str , pad_to : int ):
        ascii_bytes = bytes(astring,'ascii')
        
        return ascii_bytes + b'0x00' * (40 - len(ascii_bytes))
        
    
    
    # ------------------------------------------------------
    # Attempt to poll the queue of commands to send to quansheng
    # NOTE that now the command is also parsing the content !
    # stay here pulling commands until there are no more
    def _onePollQueue(self):
        
        while not self._oqueue.empty():
            
            # in theory there should be something, the wait time is just to be sure to never get stuck
            command : glob_ipp.Qsk5_command = self._oqueue.get(True, 0.3)
        
            self.stat.app_println("_onePollQueue: have command "+str(command))
        
            command.send_request_new(self.stat)
            
            
            

    # ------------------------------------------------------
    '''
       uint16_t w_printf:1;        // bit0 poll for printf
       uint16_t w_ipp:1;           // bit1 poll for IPP packets
       uint16_t w_dtmfipp:1;    // bit2 poll for DTMF IPP packets
       uint16_t w_counters:1;  // bit3 poll for counters
       uint16_t w_dtmfrxtx:1;  // bit4 poll for DTMF
       uint16_t w_qsmsg:1;     // bit5 poll for Quansheng messages
       uint16_t w_scanloud:1;  // bit6 poll for scanloud result
       uint16_t padding:9;     // MSB on LE CPU
    '''
    def _poll_pending(self):

        # also want spectrum_gui
        w_mask = 0b11110001

        # request the mask decided
        s_chunk = struct.pack('<H',w_mask)
        
        self.stat.libqsk5.IPP_send_chunk(IPPU_pollpending_reqid,s_chunk)
        
        r_res : RadioReceiveRes = self.stat.libqsk5.serialReceiveRes()
        
        if r_res.isIppRes():
            parser = Ippparser(r_res.pk_data)
            self.parseReceivedChunks(parser)
        elif r_res.isQuanshengRes():
            self._showQuanshengRes(r_res)
            
    # ------------------------------------------------------
    def _showQuanshengRes(self, r_res : RadioReceiveRes):

        self._println("QUANSHENG protocol")

        if r_res.error_message:
            self._println("  "+r_res.error_message)
            return
    
        bloader = BloaderBeaconParser(r_res.pk_data)
            
        self._println("  bootloader %04x %04x" % ( bloader.cmd_code, bloader.cmd_data_len) )

        if bloader.cmd_code == 0:
            return
        
        self._println("  "+bloader.b_version )
        self._println("  "+bloader.chip_id.hex() )
        
        if not self._fw_chooser_showing and not self._fw_have_write_req:
            self._runOnGuiIdle(self._clicWriteFirmware)

    
    # ------------------------------------------------------
    # If I have a firmware to load, attempt to    
    def _tryUploadFirmware(self):
        
        if not self._fw_have_write_req:
            return

        # if anything happen, I am done with this request
        self._fw_have_write_req=False

        fw_length = self.uvk5fw.getFirmwareSize()
        
        if fw_length < 50000:
            self._println("UPLOAD FIRMWARE SIZE < 50000, ABORT")
            return

        if fw_length > UVK5_FLASH_MAX_SIZE:
            self._println("UPLOAD FIRMWARE SIZE > flash size, ABORT")
            return

        libk5=self.stat.libqsk5

        # make sure there is nothing in the pipe
        libk5.serial.flush()

        qua_flash = QuaFlash(self._qconn_log, libk5)
        
        ver_bootloader : int = qua_flash.qua_FWR_wait_bootloader(10)
        
        if ver_bootloader == UVk5_BLOADER_BEACONv2_hex:
            qua_flash = Flash_V2(qua_flash)
        elif ver_bootloader == UVk5_BLOADER_BEACONv5_hex:
            qua_flash = Flash_V5(qua_flash)
        else:            
            self._println("NO BOOTLOADER MODE "+str(ver_bootloader))
            return
            
        self._println("BOOTLOADER MODE")    

        if not qua_flash.qua_FWR_TX_version():
            self._println("TX firmware version FAIL")
            return

        self._println("TX FW version OK")    
        
        # note that the fw_length is actually checked
        last_write_block_address = (fw_length & 0xFF00) + 0x100 
        
        for row in self.uvk5fw._blocklist:
            ablock : UVK5_fw_block = row
            
            if not qua_flash.qua_FWR_write_block(ablock.w_address ,last_write_block_address ,ablock.b_data): 
                self._println(" WRITE FAIL, ABORT")
                break;



    # ------------------------------------------------------
    # use this method to queue a command to be executed
    def _onePoll(self) -> bool:
        
        try:
            # do a FW upload request, if present
            self._tryUploadFirmware()
            
            # try to poll the queue of commands
            self._onePollQueue()
            
            # then do a polling for standard values
            self._poll_pending()
            
        except Exception as _err:
            self._println(traceback.format_exc())
            return False
            
        return True

    # -------------------------------------------------------------
    # @return true if it has parsed at least one chunk
    def parseReceivedChunks(self, parser : Ippparser ) -> bool:
        
        if not parser:
            return False

        have_chunks = False

        while parser.hasNextChunk():
            chunk=parser.getNextIppChunk()
            
            if chunk:
                self._showChunk(chunk)
                have_chunks=True
        
        return have_chunks
        
    # -------------------------------------------------------------
    def _showChunk(self, chunk : IppChunk ):
        
        if chunk.chunk_id == glob_ipp.IPPU_pollprintf_resid:
            self._showRadioPrintf(cast(Qsk5_res_RadioPrintf,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_pollradiostat_resid:
            self._runOnGuiIdle(self._updateRadioStatus,chunk)
        elif chunk.chunk_id == glob_ipp.IPPU_pollqsmsg_resid:
            self.stat.qsmsg_gui._println(str(chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_ident_resid:
            self._showRadioIdentity(cast(Qsk5_res_RadioIdentity,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_radiocfg_resid:
            self._showRadioConfig(cast(Qsk5_res_RadioConfiguration,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_generic_resid:
            self._showGenericResponse(cast(Qsk5_res_Generic,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_polldtmf16_resid:
            self._showPollDtmfRes(cast(Qsk5_res_PollDtmf,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_scan_loudest_resid:
            self._showScanLoudestRes(cast(Qsk5_res_Loudest,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_res_counters16:
            self._showResCounters(cast(Qsk5_res_Counters,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_res_spectrum16:
            self._showResSpectrum(cast(Qsk5_res_Spectrum,chunk))
        elif chunk.chunk_id == glob_ipp.IPPU_res_eeprom:
            self.eepromResponseFromRadio(cast(Qsk5_res_Eeprom,chunk))
        else:
            self.stat.app_println("POLL unsupported chunkid="+str(chunk.chunk_id))

    
    # ------------------------------------------------------------
    def  eepromResponseFromRadio (self, chunk : Qsk5_res_Eeprom ):
        
        if chunk.isReadResponse():
            self.stat.globeeprom.eepromUpdateFromRadio(cast(Qsk5_res_Eeprom,chunk))
        elif chunk.isWriteResponse():
            w_Address=chunk.ee_rw_address
            self.stat.eeprom_gui.setLoadingProgress(w_Address/16)

    
    
    # -------------------------------------------------------------
    # show in the main window the res counters result
    def _showResCounters(self, chunk : Qsk5_res_Counters):
        
        self._qcounters.parse_chunk(chunk)
        
            
    # -------------------------------------------------------------
    # the result should be given to the spectrum_gui window
    def _showResSpectrum(self, chunk : Qsk5_res_Spectrum):
            self._println("\nhave spectrum_gui res F= "+str(chunk.start_freq_1x))
            self.stat.spectrum_gui.showSpectrum_res(chunk)


    # -------------------------------------------------------------
    # show in the main window the identify result
    def _showScanLoudestRes(self, chunk : Qsk5_res_Loudest):
       
            self.stat.spectrum_gui._println("Scan Loudest="+str(chunk.found_F_1x))
        
        
    # -------------------------------------------------------------
    def _showPollDtmfRes(self, chunk : Qsk5_res_PollDtmf):
        self.stat.qsdtmf_gui.dtmf_println(chunk.dtmf_string)
        
    # -------------------------------------------------------------
    def _showRadioPrintf(self, chunk : Qsk5_res_RadioPrintf):
        self._qconn_log.write(chunk.print_this)
        
        
    # -------------------------------------------------------------
    # show in the main window the identify result
    def _showRadioIdentity(self, chunk : Qsk5_res_RadioIdentity):

            self.cur_radio_fw_version = chunk.sw_version
            self.cur_radio_cpu_id_hex = chunk.getCpuid_hex()
            
            self.stat.app_println("HW model="+chunk.hw_model)
            self.stat.app_println("CPU id="+self.cur_radio_cpu_id_hex)
            self.stat.app_println("SW version="+self.cur_radio_fw_version)
            
            self.stat.eeprom_gui.checkVersionCompatible(chunk.sw_version)
            self.stat.rabands_gui.update_CPU_id(self.cur_radio_cpu_id_hex)
    
    # -------------------------------------------------------------
    def _showRadioConfig(self, chunk : Qsk5_res_RadioConfiguration):
        self.stat.appmain_gui.println("RADIO CONFIG RES:")
        self.stat.appmain_gui.println("  rxf_1x "+str(chunk.rxfreq_1x))

    # -------------------------------------------------------------
    def _showGenericResponse(self, chunk : Qsk5_res_Generic):
        a_s = "GENERIC RES: oreqid "+str(chunk.from_reqid)+" reason="+str(chunk.reason)+" code="+str(chunk.rescode)+" message="+str(chunk.message)
        self.stat.appmain_gui.println(a_s)


# =========================================================================
# It is here since this is in charge of communication with radio
class Uvk5Firmware():
    
    # -------------------------------------------------------------
    def __init__(self, stat : qpystat.Qpystat):
        self._stat = stat
        self._blocklist : list[UVK5_fw_block] = []
        
    # -------------------------------------------------------------
    def _println(self, msg : str ):
        self._stat.app_println(msg)
                
    # -------------------------------------------------------------
    # You could load different FW
    def blocklist_clear(self):
        self._blocklist.clear()
        
    # -------------------------------------------------------------
    # attempt to load the given file 
    # @return true if all is fine
    def load_from_filename(self, filename ) -> bool:
        self.blocklist_clear()
        
        try:
            with open(filename, 'rb') as bin_file:
                eof=False 
                w_address=0
                
                while not eof:
                    ee_bdata = bin_file.read(UVK5_FLASH_BLOCK_LEN)
                
                    if ee_bdata:
                        self._blocklist_append(UVK5_fw_block(w_address,ee_bdata))
                
                    w_address = w_address + UVK5_FLASH_BLOCK_LEN
                    
                    if len(ee_bdata) < UVK5_FLASH_BLOCK_LEN:
                        eof=True 

            return True
                
        except Exception as exc :
            self._println("load_from_filename "+str(exc))
            return False

    # -------------------------------------------------------------
    # load all blocks using this one
    def _blocklist_append(self, a_block : UVK5_fw_block ):
        self._blocklist.append(a_block)

    # -------------------------------------------------------------
    # For verification purposes
    def getFirmwareSize(self) -> int:
        
        a_len=0;
        
        for row in self._blocklist:
            ablock : UVK5_fw_block=row
            a_len = a_len + len(ablock.b_data) 
            
        return a_len

# ----------------------------------------------------------------
class UVK5_fw_block():
    
    def __init__(self, w_address, b_data : bytes ):
        self.w_address=w_address
        self.b_data=b_data
        
        
        

